# Copyright 2024 Agnostiq Inc.

"""Unit tests for function serve decorators module."""

import pytest

import covalent_cloud as cc
from covalent_cloud.function_serve.decorators import service
from covalent_cloud.shared.classes.exceptions import InsufficientMemoryError
from covalent_cloud.shared.classes.settings import settings


def test_insufficient_memory_error():
    """Test insufficient memory error."""

    insufficient_memory = (settings.function_serve.min_executor_memory_gb - 1) * 1024
    executor = cc.CloudExecutor(
        num_cpus=4, memory=insufficient_memory, env="default", validate_environment=False
    )

    with pytest.raises(InsufficientMemoryError):
        service(executor=executor)



# Copyright 2023 Agnostiq Inc.

"""Unit tests for software environment management module."""

import os
from unittest.mock import ANY, MagicMock

from covalent_cloud.service_account_interface.auth_config_manager import AuthConfigManager
from covalent_cloud.shared.classes.settings import settings
from covalent_cloud.swe_management.models.environment import (
    BaseImageConfig,
    EnvironmentRuntimeConfig,
)
from covalent_cloud.swe_management.swe_manager import (
    create_env,
    delete_env,
    get_envs,
    get_pip_pkgs,
    missing_dependency,
    unpack_conda_pkgs,
)

TEMP_REQUIREMENT_FILEPATH = "/tmp/requirements.txt"
TEMP_ENV_YML_FILEPATH = "/tmp/environment.yml"
PIP_PKGS = [
    "numpy==1.19.5",
    "pandas==0.0.1",
    "scikit-learn",
    "aiohttp>1.0.0",
    "boto3<=1.0.0",
    "matplotlib<=1.0.0,>.0.3.4",
    "request!=1.0.0",
    "covalent-cloud>=0.61.0rc0",
]
CONDA_ENV_YML = """
name: mockenv
channels:
  - javascript
dependencies:
  - python=3.9
  - bokeh=2.4.2
  - conda-forge::numpy=1.21.*
  - nodejs=16.13.*
  - flask
  - pip
  - pip:
    - Flask-Testing
"""
CONDA_CHANNELS = ["javascript"]
CONDA_DEPENDENCIES = [
    "python=3.9",
    "bokeh=2.4.2",
    "conda-forge::numpy=1.21.*",
    "nodejs=16.13.*",
    "flask",
    "pip",
    {"pip": ["Flask-Testing"]},
]
MOCK_CONDA_ENV_NAME = "mockenv"
MOCK_PIP_PKGS = ["mock-package-1", "mock-package-2"]
MOCK_CHANNELS = ["mock-channel-1", "mock-channel-2"]
MOCK_DEPENDENCIES = [
    "mock-dependency-1",
    "mock-dependency-2",
    {"pip": ["mock-pip-dependency-1", "mock-pip-dependency-2"]},
]

TEST_API_KEY = "fake_api_key"  # pragma: allowlist secret


def create_temp_requirement_file():
    """Create a temporary requirements.txt file."""
    with open(TEMP_REQUIREMENT_FILEPATH, "w") as f:
        for pkg in PIP_PKGS:
            f.write(pkg + "\n")


def remove_temp_requirement_file():
    """Remove the temporary requirements.txt file."""
    os.remove(TEMP_REQUIREMENT_FILEPATH)


def create_conda_env_file():
    """Create a temporary conda environment.yml file."""
    with open(TEMP_ENV_YML_FILEPATH, "w") as f:
        f.write(CONDA_ENV_YML)


def remove_conda_env_file():
    """Remove the temporary conda environment.yml file."""
    os.remove(TEMP_ENV_YML_FILEPATH)


def test_get_pip_pkgs_str():
    """Test the get pip packages function."""
    create_temp_requirement_file()
    pip_pkgs = get_pip_pkgs(TEMP_REQUIREMENT_FILEPATH)
    assert pip_pkgs == PIP_PKGS
    remove_temp_requirement_file()


def test_get_pip_pkgs_list():
    """Test the get pip packages function."""
    create_temp_requirement_file()
    pip_pkgs = get_pip_pkgs(PIP_PKGS)
    assert pip_pkgs == PIP_PKGS
    remove_temp_requirement_file()


def test_get_pip_pkgs_list_with_requirements():
    """Test the get pip packages function."""
    create_temp_requirement_file()
    pip_pkgs = get_pip_pkgs(["mock-package", TEMP_REQUIREMENT_FILEPATH, "mock-package-2"])
    assert pip_pkgs == ["mock-package"] + PIP_PKGS + ["mock-package-2"]
    remove_temp_requirement_file()


def test_unpack_conda_pkgs_str():
    """Test the unpack conda packages function."""
    create_conda_env_file()
    channels, dependencies = unpack_conda_pkgs(TEMP_ENV_YML_FILEPATH)
    assert channels == CONDA_CHANNELS
    assert dependencies == CONDA_DEPENDENCIES
    remove_conda_env_file()


def test_unpack_conda_pkgs_list():
    """Test the unpack conda packages function."""
    create_conda_env_file()
    channels, dependencies = unpack_conda_pkgs(CONDA_DEPENDENCIES)
    assert channels == []
    assert dependencies == CONDA_DEPENDENCIES
    remove_conda_env_file()


def test_unpack_conda_pkgs_dict():
    """Test the unpack conda packages function."""
    create_conda_env_file()
    channels, dependencies = unpack_conda_pkgs(
        {
            "channels": CONDA_CHANNELS,
            "dependencies": CONDA_DEPENDENCIES,
        }
    )
    assert channels == CONDA_CHANNELS
    assert dependencies == CONDA_DEPENDENCIES
    remove_conda_env_file()


def test_create_with_base_image(mocker):
    """Test the create environment function."""
    MOCK_CREATE_DEPENDENCIES = ["mock-dependency-1", "mock-dependency-2"]
    AuthConfigManager.save_api_key(TEST_API_KEY)  # This will overite your local key

    session_mock = MagicMock()
    mocker.patch("requests.Session.__enter__", return_value=session_mock)

    mock_file_handle = MagicMock()
    mock_file_handle.name = "mock_name"
    mock_ctx_mgr = MagicMock()
    mock_ctx_mgr.__enter__ = MagicMock(return_value=mock_file_handle)

    mocker.patch("covalent_cloud.swe_management.swe_manager.open", return_value=mock_ctx_mgr)

    get_pip_pkgs_mock = mocker.patch(
        "covalent_cloud.swe_management.swe_manager.get_pip_pkgs", return_value=MOCK_PIP_PKGS
    )
    unpack_conda_pkgs_mock = mocker.patch(
        "covalent_cloud.swe_management.swe_manager.unpack_conda_pkgs",
        return_value=(MOCK_CHANNELS, MOCK_CREATE_DEPENDENCIES),
    )

    create_env(
        name=MOCK_CONDA_ENV_NAME,
        pip=MOCK_PIP_PKGS,
        conda=MOCK_CREATE_DEPENDENCIES,
        base_image="the_magical_land_of_narnia",
    )

    get_pip_pkgs_mock.assert_called_once_with(MOCK_PIP_PKGS)
    unpack_conda_pkgs_mock.assert_called_once_with(MOCK_CREATE_DEPENDENCIES)

    config = EnvironmentRuntimeConfig(
        image=BaseImageConfig(base_image="the_magical_land_of_narnia")
    )
    data = {
        "name": MOCK_CONDA_ENV_NAME,
        "runtime_config": config.model_dump_json(exclude_none=True),
    }
    files = {"definition": mock_file_handle}
    session_mock.post.assert_called_once_with(
        f"{settings.dispatcher_uri}/api/v2/envs", data=data, files=files, headers=ANY
    )


def test_create_env_empty(mocker):
    """Test the create environment function."""
    AuthConfigManager.save_api_key(TEST_API_KEY)  # This will overite your local key

    temp_file_mock = mocker.patch("tempfile.NamedTemporaryFile")
    session_mock = MagicMock()
    mocker.patch("requests.Session.__enter__", return_value=session_mock)

    mock_file_handle = MagicMock()
    mock_file_handle.name = "mock_name"
    mock_ctx_mgr = MagicMock()
    mock_ctx_mgr.__enter__ = MagicMock(return_value=mock_file_handle)

    mocker.patch("covalent_cloud.swe_management.swe_manager.open", return_value=mock_ctx_mgr)

    create_env(name=MOCK_CONDA_ENV_NAME)  # no `pip` or `conda` dependencies

    data = {
        "name": MOCK_CONDA_ENV_NAME,
    }
    files = {"definition": mock_file_handle}
    session_mock.post.assert_called_once_with(
        f"{settings.dispatcher_uri}/api/v2/envs", data=data, files=files, headers=ANY
    )


def test_create_env(mocker):
    """Test the create environment function."""
    MOCK_CREATE_DEPENDENCIES = ["mock-dependency-1", "mock-dependency-2"]
    AuthConfigManager.save_api_key(TEST_API_KEY)  # This will overite your local key

    temp_file_mock = mocker.patch("tempfile.NamedTemporaryFile")
    session_mock = MagicMock()
    mocker.patch("requests.Session.__enter__", return_value=session_mock)

    mock_file_handle = MagicMock()
    mock_file_handle.name = "mock_name"
    mock_ctx_mgr = MagicMock()
    mock_ctx_mgr.__enter__ = MagicMock(return_value=mock_file_handle)

    mocker.patch("covalent_cloud.swe_management.swe_manager.open", return_value=mock_ctx_mgr)

    get_pip_pkgs_mock = mocker.patch(
        "covalent_cloud.swe_management.swe_manager.get_pip_pkgs", return_value=MOCK_PIP_PKGS
    )
    unpack_conda_pkgs_mock = mocker.patch(
        "covalent_cloud.swe_management.swe_manager.unpack_conda_pkgs",
        return_value=(MOCK_CHANNELS, MOCK_CREATE_DEPENDENCIES),
    )

    create_env(
        name=MOCK_CONDA_ENV_NAME,
        pip=MOCK_PIP_PKGS,
        conda=MOCK_CREATE_DEPENDENCIES,
    )
    get_pip_pkgs_mock.assert_called_once_with(MOCK_PIP_PKGS)
    unpack_conda_pkgs_mock.assert_called_once_with(MOCK_CREATE_DEPENDENCIES)

    data = {
        "name": MOCK_CONDA_ENV_NAME,
    }
    files = {"definition": mock_file_handle}
    session_mock.post.assert_called_once_with(
        f"{settings.dispatcher_uri}/api/v2/envs", data=data, files=files, headers=ANY
    )


def test_get_envs(mocker):
    """Test the get environments function."""
    mock_get_envs_filtered = mocker.patch(
        "covalent_cloud.swe_management.swe_manager._get_envs_filtered"
    )
    mock_get_envs_filtered.return_value = {"mock-env-name": "mock-env"}

    envs = get_envs()

    mock_get_envs_filtered.assert_called_once_with(count=10, status="READY")
    assert envs == {"mock-env-name": "mock-env"}


def test_delete_env(mocker):
    """Test the delete environment function."""

    env_mock = MagicMock()
    mock_env_name = "mock-env-name"
    env_mock.name = mock_env_name

    mock_client = mocker.patch("covalent_cloud.swe_management.swe_manager.get_client")

    mock_get_envs_filtered = mocker.patch(
        "covalent_cloud.swe_management.swe_manager._get_envs_filtered"
    )
    mock_get_envs_filtered.return_value = {mock_env_name: env_mock}

    delete_env(mock_env_name)

    mock_get_envs_filtered.assert_called_once_with(name="mock-env-name")
    mock_client.return_value.delete.assert_called_once_with(f"/api/v2/envs/{env_mock.id}")


def test_missing_dependency_returns_none_when_missing():
    """Test the dependency is missing function."""
    dependencies = ["mock-dependency-1", "mock-dependency-2", "python-cpp-dep", "llm-cpp-python"]
    assert missing_dependency("python=", dependencies) is None


def test_missing_dependency_returns_dep_when_not_missing():
    """Test the dependency is missing function."""
    dependencies = ["mock-dependency-1", "python=3.9", "mock-dependency-2"]
    assert missing_dependency("python", dependencies) == "python=3.9"


def test_missing_dependency_returns_none_when_exact_match():
    """Test the dependency is missing function."""
    dependencies = ["mock-dependency-1", "pip:"]
    assert missing_dependency("pip", dependencies, True) is None


def test_missing_dependency_returns_dep_when_not_missing_with_exact():
    """Test the dependency is missing function."""
    dependencies = ["python=3.10", "pip", "pip:"]
    assert missing_dependency("pip", dependencies, True) == "pip"



# Copyright 2023 Agnostiq Inc.


from importlib import metadata

from . import cloud_executor
from .cloud_executor.cloud_executor import CloudExecutor
from .cloud_executor.oci_cloud_executor import OCICloudExecutor
from .dispatch_management import cancel, dispatch, get_result, redispatch
from .function_serve.decorators import service
from .function_serve.deployment import deploy, get_deployment
from .function_serve.models import Deployment
from .qelectron_sdk.executors import CloudQCluster
from .service_account_interface.auth_config_manager import get_api_key, save_api_key
from .service_account_interface.client import get_client
from .shared.classes.settings import settings
from .swe_management.secrets_manager import delete_secret, list_secrets, store_secret
from .swe_management.swe_manager import create_env, delete_env, get_envs
from .volume.volume import volume

__version__ = metadata.version("covalent_cloud")

__all__ = [
    "cloud_executor",
    "CloudExecutor",
    "OCICloudExecutor",
    "CloudQCluster",
    "cancel",
    "dispatch",
    "get_result",
    "redispatch",
    "get_api_key",
    "save_api_key",
    "get_client",
    "settings",
    "delete_secret",
    "list_secrets",
    "store_secret",
    "create_env",
    "get_envs",
    "delete_env",
    "volume",
    "deploy",
    "get_deployment",
    "Deployment",
    "service",
]



# Copyright 2024 Agnostiq Inc.

import inspect
from functools import wraps
from typing import Callable, Optional

from covalent_cloud.cloud_executor.cloud_executor import CloudExecutor
from covalent_cloud.function_serve.common import SupportedMethods
from covalent_cloud.function_serve.service_class import FunctionService
from covalent_cloud.shared.classes.exceptions import InsufficientMemoryError
from covalent_cloud.shared.classes.settings import settings
from covalent_cloud.volume.volume import Volume

__all__ = [
    "service",
]


def service(
    _func: Optional[Callable] = None,
    *,
    executor: CloudExecutor,
    name: str = "",
    description: str = "",
    auth: Optional[bool] = None,
    tags: Optional[list] = None,
    compute_share: float = 1.0,
    volume: Optional[Volume] = None,
) -> FunctionService:
    """Decorator used to create a function service by wrapping its initializer.

    Args:
        _func: Service initializer function to be wrapped. This function should return a
            dict-like object containing initialized values to be used by the service's endpoints.
        executor: A CloudExecutor instance that specifies compute resources assigned to this service.
        name: Optional name for the function service. Defaults to the decorated function's name.
        description: Optional description for the function service.
            Defaults to the decorated function's docstring.
        auth: Whether or not to require an authentication token for requests.
            Auth is required by default.
        tags: An optional list of tags to associate with the function service. Defaults to None.
        compute_share: Fraction of resources to allocate to each replica of the function service.
            Defaults to 1, meaning a single replica that uses all assigned resources.
        volume: Choice of cloud storage volume to attach. Defaults to None.

    Raises:
        InsufficientMemoryError: If the assigned executor has less than the minimum required memory.

    Returns:
        A function service object with an `endpoint` decorator method for creating HTTP routes.
    """
    if executor.memory < settings.function_serve.min_executor_memory_gb * 1024:
        raise InsufficientMemoryError(int(executor.memory / 1024))

    tags = tags or []

    def service_decorator(func=None):
        @wraps(func)
        def internal_wrapper(executor, name, description, auth, tags, compute_share, volume):
            return FunctionService(
                func,
                executor,
                name,
                description,
                auth,
                tags,
                compute_share,
                volume,
                _main_func=func,
            )

        return internal_wrapper(executor, name, description, auth, tags, compute_share, volume)

    if _func is None:
        return service_decorator

    return service_decorator(_func)


def op_internal_wrapper_generator(
    route,
    executor,
    name,
    streaming,
    description,
    auth,
    tags,
    compute_share,
    op,
):
    fs_instance = service(
        executor=executor,
        name=name,
        description=description,
        auth=auth,
        tags=tags,
        compute_share=compute_share,
    )(None)

    # Get the correct method from the FunctionService class, e.g. _get, _post, _put, etc.
    op_func = getattr(fs_instance, "_" + op.value.lower())

    def internal_wrapper(func):

        # Set the function service instance attributes since there is no init_func function
        fs_instance.func_name = func.__name__
        fs_instance.name = fs_instance.name or func.__name__
        fs_instance.func_source = inspect.getsource(func)
        fs_instance.func_description = description or func.__doc__ or ""
        fs_instance._main_func = func

        # Register the route and the method function and return the FunctionService instance
        return op_func(route, name, description, streaming)(func)

    return internal_wrapper


def get_decorator(
    route,
    executor=None,
    name=None,
    streaming=False,
    description=None,
    auth=None,
    tags=None,
    compute_share=None,
):
    return op_internal_wrapper_generator(
        route,
        executor,
        name,
        streaming,
        description,
        auth,
        tags,
        compute_share,
        SupportedMethods.GET,
    )


def post_decorator(
    route,
    executor=None,
    name=None,
    streaming=False,
    description=None,
    auth=None,
    tags=None,
    compute_share=None,
):

    return op_internal_wrapper_generator(
        route,
        executor,
        name,
        streaming,
        description,
        auth,
        tags,
        compute_share,
        SupportedMethods.POST,
    )


def put_decorator(
    route,
    executor=None,
    name=None,
    streaming=False,
    description=None,
    auth=None,
    tags=None,
    compute_share=None,
):

    return op_internal_wrapper_generator(
        route,
        executor,
        name,
        streaming,
        description,
        auth,
        tags,
        compute_share,
        SupportedMethods.PUT,
    )


def delete_decorator(
    route,
    executor=None,
    name=None,
    streaming=False,
    description=None,
    auth=None,
    tags=None,
    compute_share=None,
):

    return op_internal_wrapper_generator(
        route,
        executor,
        name,
        streaming,
        description,
        auth,
        tags,
        compute_share,
        SupportedMethods.DELETE,
    )


def patch_decorator(
    route,
    executor=None,
    name=None,
    streaming=False,
    description=None,
    auth=None,
    tags=None,
    compute_share=None,
):

    return op_internal_wrapper_generator(
        route,
        executor,
        name,
        streaming,
        description,
        auth,
        tags,
        compute_share,
        SupportedMethods.PATCH,
    )


# Assigning the decorators to the service function

# We still do this instead of directly using `op_internal_wrapper_generator`
# so that the correct supported arguments are shown when something like autocompletion is used
service._get = get_decorator
service._post = post_decorator
service._put = put_decorator
service._delete = delete_decorator

# This is the one that will be used by the user
service.endpoint = post_decorator



# Copyright 2024 Agnostiq Inc.

from concurrent.futures import ThreadPoolExecutor, wait
from typing import Generic, List, TypeVar

import covalent as ct
import requests
from pydantic import BaseModel
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from covalent_cloud.function_serve.common import ServeAssetType
from covalent_cloud.function_serve.models import ServeAsset
from covalent_cloud.service_account_interface.client import get_deployment_client
from covalent_cloud.shared.classes.settings import Settings, settings

_upload_executor = ThreadPoolExecutor()


ModelType = TypeVar("ModelType", bound=BaseModel)


class AssetsMediator(Generic[ModelType]):

    serve_assets: List[ServeAsset] = []

    def __init__(self) -> None:
        self.serve_assets = []

    def hydrate_assets_from_model(self, model, settings: Settings = settings) -> ModelType:
        """
        Add serve asset ids and pre-signed urls to any pydantic model with ServeAsset fields.

        Args:
            model: A pydantic model that has ServeAsset fields.

        Returns:
            A new instance of the model with ServeAsset fields populated with ids and pre-signed urls.
        """
        _model_dump = model.model_dump()
        deployment_client = get_deployment_client(settings)

        def find_and_replace_serialized_assets(data):
            if isinstance(data, dict):

                if (type_ := data.get("type")) and type_ in [
                    ServeAssetType.ASSET,
                    ServeAssetType.JSON,
                ]:
                    serve_asset = ServeAsset(**data)
                    self.serve_assets.append(serve_asset)
                    return serve_asset

                return {
                    key: find_and_replace_serialized_assets(value) for key, value in data.items()
                }

            if isinstance(data, list):
                return [find_and_replace_serialized_assets(item) for item in data]

            return data

        new_schema = find_and_replace_serialized_assets(_model_dump)

        res = deployment_client.post(
            "/assets", request_options={"params": {"n": len(self.serve_assets)}}
        )
        presigned_assets = res.json()

        for serve_asset in self.serve_assets:
            asset = presigned_assets.pop()
            serve_asset.url = asset.get("url")
            serve_asset.id = asset.get("id")

        return type(model)(**new_schema)

    def upload_all(self):
        """
        Upload all ServeAssets to the cloud.
        """
        _upload_futures = []

        serve_assets = self.serve_assets
        for serve_asset in serve_assets:
            fut = _upload_executor.submit(AssetsMediator.upload_asset, serve_asset)
            _upload_futures.append(fut)

        done_futures, not_done_futures = wait(_upload_futures)

        self.serve_assets = []

        # ensure all futures are done and raise any exceptions if any
        for fut in done_futures:
            try:
                fut.result()
            except Exception as e:
                raise e

        for fut in not_done_futures:
            try:
                fut.result()
            except Exception as e:
                raise e

    @staticmethod
    def upload_asset(asset: ServeAsset, settings: Settings = settings) -> None:
        url = asset.url

        if url is None:
            deployment_client = get_deployment_client(settings)
            res = deployment_client.post("/assets")
            presigned_urls = res.json()
            asset.url = presigned_urls[0]["url"]
            asset.id = presigned_urls[0]["id"]

        if asset.type == ServeAssetType.ASSET:
            transportable_object: ct.TransportableObject = ct.TransportableObject.deserialize(
                asset.serialized_object
            )
            data: bytes = transportable_object.serialize()

        elif asset.type == ServeAssetType.JSON:
            # No need to pickle this.
            data: bytes = asset.serialized_object

        else:
            raise ValueError(f"Unsupported asset type: '{asset.type}'")

        retry_strategy = Retry(
            total=5,
            backoff_factor=0.1,
            allowed_methods={"PUT"},
        )
        session = requests.Session()
        session.mount("https://", HTTPAdapter(max_retries=retry_strategy))
        session.mount("http://", HTTPAdapter(max_retries=retry_strategy))
        r = session.put(asset.url, data=data)
        r.raise_for_status()

        asset.url = None
        asset.serialized_object = None



# Copyright 2023 Agnostiq Inc.


"""Helpers for result fetching."""


from __future__ import annotations

import enum
import os
from pathlib import Path
from typing import Any, Optional

import requests
from covalent._results_manager.result import Result
from covalent._results_manager.results_manager import SDK_LAT_META_KEYS, SDK_NODE_META_KEYS
from covalent._results_manager.wait import EXTREME
from covalent._serialize.common import load_asset
from covalent._serialize.electron import ASSET_FILENAME_MAP as ELECTRON_ASSET_FILENAMES
from covalent._serialize.electron import ASSET_TYPES as ELECTRON_ASSET_TYPES
from covalent._serialize.lattice import ASSET_FILENAME_MAP as LATTICE_ASSET_FILENAMES
from covalent._serialize.lattice import ASSET_TYPES as LATTICE_ASSET_TYPES
from covalent._serialize.result import ASSET_FILENAME_MAP as RESULT_ASSET_FILENAMES
from covalent._serialize.result import ASSET_TYPES as RESULT_ASSET_TYPES
from covalent._serialize.result import deserialize_result
from covalent._shared_files import logger
from covalent._shared_files.exceptions import MissingLatticeRecordError
from covalent._shared_files.schemas.asset import AssetSchema
from covalent._shared_files.schemas.result import ResultSchema
from covalent._workflow.transport import TransportableObject
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from ..shared.classes.api import AuthConfigManager
from ..shared.classes.settings import Settings, settings

app_log = logger.app_log
log_stack_info = logger.log_stack_info


# Multi-part


class AssetScope(enum.Enum):
    RESULT = "result"
    LATTICE = "lattice"
    NODE = "node"


# Utility function to decode TransportableObjects
def _decode(value: Any):
    if isinstance(value, TransportableObject):
        return value.get_deserialized()
    else:
        return value


class FutureVar:
    """A class that represents a variable that is not yet available.

    List of private attributes: _dispatch_id, _task_id, _name
    List of public and immutable attributes: value
    """

    def __init__(
        self,
        result_manager: ResultManager,
        scope: AssetScope,
        name: str,
        dispatch_id: str,
        task_id: Optional[int],
    ):
        self._scope = scope
        self._dispatch_id = dispatch_id
        self._task_id = task_id
        self._name = name
        self._value = None
        self._result_manager = result_manager
        self._downloaded = False

        if self._scope == AssetScope.RESULT:
            size = self._result_manager._manifest["assets"][name]["size"]

        if self._scope == AssetScope.LATTICE:
            size = self._result_manager._manifest["lattice"]["assets"][name]["size"]

        if self._scope == AssetScope.NODE:
            node = self._result_manager._manifest["lattice"]["transport_graph"]["nodes"][task_id]
            size = node["assets"][name]["size"]

        self._size = size

    @property
    def value(self):
        return self._value

    @property
    def size(self):
        return self._size

    # TODO - save artifacts to local storage first instead of loading
    # them entirely in memory.
    def load(self) -> Any:
        """Query Covalent Cloud to retrieve and populate 'value' attribute.'"""

        if self._scope == AssetScope.RESULT:
            if not self._downloaded:
                self._result_manager.download_result_asset(self._name)
                self._downloaded = True
            self._value = _decode(self._result_manager.load_result_asset(self._name))

        if self._scope == AssetScope.LATTICE:
            if not self._downloaded:
                self._result_manager.download_lattice_asset(self._name)
                self._downloaded = True
            self._value = _decode(self._result_manager.load_lattice_asset(self._name))

        if self._scope == AssetScope.NODE:
            if not self._downloaded:
                self._result_manager.download_node_asset(self._task_id, self._name)
                self._downloaded = True
            self._value = _decode(self._result_manager.load_node_asset(self._task_id, self._name))

        return self._value

    def __str__(self) -> str:
        if self._downloaded:
            return str(self.value)
        else:
            return "<FutureVar>"

    # infinite recursion
    def __deepcopy__(self, memo):
        """Make a deep copy.

        Detaches the result_manager attribute to avoid infinite
        recursion.

        """
        cls = type(self)
        fv = cls.__new__(cls)
        fv.__init__(None, self._scope, self._name, self._dispatch_id, self._task_id)
        return fv


class FutureResult(Result):
    """A class that represents a dispatch result that is not yet available.

    List of private attributes: _dispatch_id, _root_dispatch_id
    List of public and immutable attributes: dispatch_name, start_time, end_time, status, lattice, result, inputs, error
    """

    @property
    def result(self) -> Any:
        return self._result

    @property
    def inputs(self) -> Any:
        return super().inputs

    @property
    def error(self) -> str:
        return self._error

    @staticmethod
    def _from_result_object(res: Result):
        fr = FutureResult(res.lattice)
        fr.__dict__ = res.__dict__
        return fr

    def __str__(self):
        """String representation of the result object"""

        show_result_str = f"""
Lattice Result
==============
status: {self._status}
result: {self.result}
inputs: {self.inputs}
error: {self.error}

start_time: {self.start_time}
end_time: {self.end_time}

dispatch_id: {self.dispatch_id}

Node Outputs
------------
"""
        node_outputs = self.get_all_node_outputs()
        for k, v in node_outputs.items():
            show_result_str += f"{k}: {v}\n"

        return show_result_str


def _get_result_export_from_dispatcher(
    dispatch_id: str,
    wait: bool = False,
    status_only: bool = False,
    settings: Settings = settings,
) -> ResultSchema:
    """
    Internal function to get the results of a dispatch from the server without checking if it is ready to read.

    Args:
        dispatch_id: The dispatch id of the result.
        wait: Controls how long the method waits for the server to return a result. If False, the method will not wait and will return the current status of the workflow. If True, the method will wait for the result to finish and keep retrying for sys.maxsize.
        status_only: If true, only returns result status, not the full result object, default is False.
        dispatcher_addr: Dispatcher server address, defaults to the address set in covalent.config.

    Returns:
        The result object from the server.

    Raises:
        MissingLatticeRecordError: If the result is not found.
    """

    dispatcher_addr = settings.dispatcher_uri
    dispatcher_port = settings.dispatcher_port

    dispatcher_uri = dispatcher_addr

    if dispatcher_port is not None:
        dispatcher_uri += f":{dispatcher_port}"

    retries = int(EXTREME) if wait else 5

    adapter = HTTPAdapter(max_retries=Retry(total=retries, backoff_factor=1))
    http = requests.Session()
    http.mount("http://", adapter)
    http.mount("https://", adapter)
    url = dispatcher_uri + "/api/v1/export/" + dispatch_id

    api_key = AuthConfigManager.get_api_key(settings)
    headers = {"x-api-key": api_key}

    response = http.get(
        url,
        params={"wait": bool(int(wait)), "status_only": status_only},
        headers=headers,
    )
    if response.status_code == 404:
        raise MissingLatticeRecordError
    response.raise_for_status()
    export = response.json()
    return export


# Functions for computing local URIs
def get_node_asset_path(results_dir: str, node_id: int, key: str):
    filename = ELECTRON_ASSET_FILENAMES[key]
    return results_dir + f"/node_{node_id}/{filename}"


def get_lattice_asset_path(results_dir: str, key: str):
    filename = LATTICE_ASSET_FILENAMES[key]
    return results_dir + f"/{filename}"


def get_result_asset_path(results_dir: str, key: str):
    filename = RESULT_ASSET_FILENAMES[key]
    return results_dir + f"/{filename}"


# Asset transfers


def download_asset(remote_uri: str, local_path: str, chunk_size: int = 1024 * 1024):
    api_key = AuthConfigManager.get_api_key(settings)
    headers = {"x-api-key": api_key}
    r = requests.get(remote_uri, stream=True, headers=headers)
    # Missing objects correspond to `None` attributes

    if r.status_code == 404:
        return None
    r.raise_for_status()
    with open(local_path, "wb") as f:
        for chunk in r.iter_content(chunk_size=chunk_size):
            f.write(chunk)
    return f"file://{local_path}"


def _get_result_asset_uri(manifest: dict, key: str) -> str:
    return manifest["assets"][key]["remote_uri"]


def _download_result_asset(manifest: dict, results_dir: str, key: str):
    remote_uri = manifest["assets"][key]["remote_uri"]
    local_path = get_result_asset_path(results_dir, key)
    manifest["assets"][key]["uri"] = download_asset(remote_uri, local_path)


def _get_lattice_asset_uri(manifest: dict, key: str) -> str:
    lattice_assets = manifest["lattice"]["assets"]
    return lattice_assets[key]["remote_uri"]


def _download_lattice_asset(manifest: dict, results_dir: str, key: str):
    lattice_assets = manifest["lattice"]["assets"]
    remote_uri = lattice_assets[key]["remote_uri"]
    local_path = get_lattice_asset_path(results_dir, key)
    download_asset(remote_uri, local_path)
    lattice_assets[key]["uri"] = download_asset(remote_uri, local_path)


def _get_node_asset_uri(manifest: dict, node_id: int, key: str):
    node = manifest["lattice"]["transport_graph"]["nodes"][node_id]
    node_assets = node["assets"]
    return node_assets[key]["remote_uri"]


def _download_node_asset(manifest: dict, results_dir: str, node_id: int, key: str):
    node = manifest["lattice"]["transport_graph"]["nodes"][node_id]
    node_assets = node["assets"]
    remote_uri = node_assets[key]["remote_uri"]
    local_path = get_node_asset_path(results_dir, node_id, key)
    node_assets[key]["uri"] = download_asset(remote_uri, local_path)


def _load_result_asset(manifest: dict, key: str):
    asset_meta = AssetSchema(**manifest["assets"][key])
    return load_asset(asset_meta, RESULT_ASSET_TYPES[key])


def _load_lattice_asset(manifest: dict, key: str):
    asset_meta = AssetSchema(**manifest["lattice"]["assets"][key])
    return load_asset(asset_meta, LATTICE_ASSET_TYPES[key])


def _load_node_asset(manifest: dict, node_id: int, key: str):
    node = manifest["lattice"]["transport_graph"]["nodes"][node_id]
    asset_meta = AssetSchema(**node["assets"][key])
    return load_asset(asset_meta, ELECTRON_ASSET_TYPES[key])


class ResultManager:
    def __init__(self, manifest: ResultSchema, results_dir: str):
        self.result_object = deserialize_result(manifest)
        self._manifest = manifest.dict()
        self._results_dir = results_dir

    def save(self, path: Optional[str] = None):
        if not path:
            path = os.path.join(self._results_dir, "manifest.json")
        with open(path, "w") as f:
            f.write(ResultSchema.parse_obj(self._manifest).json())

    @staticmethod
    def load(path: str, results_dir: str) -> "ResultManager":
        with open(path, "r") as f:
            manifest_json = f.read()

        return ResultManager(ResultSchema.parse_raw(manifest_json), results_dir)

    def _populate_result_object(self):
        dispatch_id = self._manifest["metadata"]["dispatch_id"]

        # result assets
        for key in RESULT_ASSET_TYPES:
            remote_uri = _get_result_asset_uri(self._manifest, key)
            fv = FutureVar(self, AssetScope.RESULT, key, dispatch_id, None)
            self.result_object.__dict__[f"_{key}"] = fv

        # lattice assets
        for key in LATTICE_ASSET_TYPES:
            remote_uri = _get_lattice_asset_uri(self._manifest, key)
            fv = FutureVar(self, AssetScope.LATTICE, key, dispatch_id, None)
            if key in SDK_LAT_META_KEYS:
                self.result_object.lattice.metadata[key] = fv
            else:
                self.result_object.lattice.__dict__[key] = fv

        # transport graph node assets
        tg = self.result_object.lattice.transport_graph
        for node_id in tg._graph.nodes:
            for key in ELECTRON_ASSET_TYPES:
                remote_uri = _get_node_asset_uri(self._manifest, node_id, key)
                fv = FutureVar(self, AssetScope.NODE, key, dispatch_id, node_id)
                if key in SDK_NODE_META_KEYS:
                    node_meta = tg.get_node_value(node_id, "metadata")
                    node_meta[key] = fv
                else:
                    tg.set_node_value(node_id, key, fv)

    def download_result_asset(self, key: str):
        _download_result_asset(self._manifest, self._results_dir, key)

    def download_lattice_asset(self, key: str):
        _download_lattice_asset(self._manifest, self._results_dir, key)

    def download_node_asset(self, node_id: int, key: str):
        _download_node_asset(self._manifest, self._results_dir, node_id, key)

    def load_result_asset(self, key: str) -> Any:
        return _load_result_asset(self._manifest, key)

    def load_lattice_asset(self, key: str) -> Any:
        return _load_lattice_asset(self._manifest, key)

    def load_node_asset(self, node_id: int, key: str) -> Any:
        return _load_node_asset(self._manifest, node_id, key)

    @staticmethod
    def from_dispatch_id(
        dispatch_id: str,
        results_dir: str,
        wait: bool = False,
        settings: Settings = settings,
    ) -> "ResultManager":
        export = _get_result_export_from_dispatcher(
            dispatch_id,
            wait,
            status_only=False,
            settings=settings,
        )

        manifest = ResultSchema.parse_obj(export["result_export"])

        new_electron_schemas = []
        for electron_schema in manifest.lattice.transport_graph.nodes:
            electron_schema.assets.qelectron_db = AssetSchema(size=0)
            new_electron_schemas.append(electron_schema)

        manifest.lattice.transport_graph.nodes = new_electron_schemas

        # sort the nodes
        manifest.lattice.transport_graph.nodes.sort(key=lambda x: x.id)

        rm = ResultManager(manifest, results_dir)
        result_object = rm.result_object
        result_object._results_dir = results_dir
        Path(results_dir).mkdir(parents=True, exist_ok=True)

        # Create node subdirectories
        for node_id in result_object.lattice.transport_graph._graph.nodes:
            node_dir = results_dir + f"/node_{node_id}"
            Path(node_dir).mkdir(exist_ok=True)

        return rm


def get_result_manager(dispatch_id, results_dir=None, wait=False, settings=settings):
    if not results_dir:
        results_dir = settings.results_dir + f"/{dispatch_id}"
    return ResultManager.from_dispatch_id(dispatch_id, results_dir, wait, settings)


def _get_result_multistage(
    dispatch_id: str,
    wait: bool = False,
    settings: Settings = settings,
    status_only: bool = False,
) -> Result:
    """
    Get the results of a dispatch from a file.

    Args:
        dispatch_id: The dispatch id of the result.
        wait: Controls how long the method waits for the server to return a result. If False, the method will not wait and will return the current status of the workflow. If True, the method will wait for the result to finish and keep retrying for sys.maxsize.

    Returns:
        The result from the file.

    """

    try:
        if status_only:
            return _get_result_export_from_dispatcher(
                dispatch_id=dispatch_id,
                wait=wait,
                status_only=status_only,
                settings=settings,
            )
        else:
            rm = get_result_manager(dispatch_id, None, wait, settings)
            rm._populate_result_object()

    except MissingLatticeRecordError as ex:
        app_log.warning(
            f"Dispatch ID {dispatch_id} was not found in the database. Incorrect dispatch id."
        )

        raise ex

    return rm.result_object


def get_result(
    dispatch_id: str,
    wait: bool = False,
    settings: Settings = settings,
    status_only: bool = False,
) -> Result:

    if not dispatch_id:
        return None

    res = _get_result_multistage(
        dispatch_id=dispatch_id,
        wait=wait,
        settings=settings,
        status_only=status_only,
    )
    if status_only:
        return res
    else:
        result_object = res

    # Populate sublattice results
    tg = result_object.lattice.transport_graph
    for node_id in tg._graph.nodes:
        sub_dispatch_id = tg.get_node_value(node_id, "sub_dispatch_id")
        sub_result_object = get_result(sub_dispatch_id, wait, settings, status_only)
        tg.set_node_value(node_id, "sublattice_result", sub_result_object)

    fr = FutureResult._from_result_object(result_object)

    return fr



# Copyright 2023 Agnostiq Inc.

from unittest.mock import MagicMock
from covalent_cloud.shared.classes.api import APIClient
from covalent_cloud.shared.classes.settings import AuthSettings, Settings


class TestAPIClient:

    MOCK_URI = "https://www.example.com"

    def test_get_global_headers(self, mocker):
        MOCK_API_KEY = "mock-api-key"  # pragma: allowlist secret
        # substituting token avoids fetching/creating from config file
        settings = Settings(auth=AuthSettings(api_key=MOCK_API_KEY))
        MOCK_GLOBAL_HEADERS = {
            "x-api-key": MOCK_API_KEY,
        }
        api = APIClient(host_uri=self.MOCK_URI, headers={"CLIENT": "TEST"}, settings=settings)
        assert api.get_global_headers() == {"CLIENT": "TEST", **MOCK_GLOBAL_HEADERS}

    def test_prepare_request(self, mocker):
        # substituting token avoids fetching/creating from config file
        MOCK_API_KEY = "mock_api_key"  # pragma: allowlist secret
        settings = Settings(auth=AuthSettings(api_key=MOCK_API_KEY))
        api = APIClient(host_uri="http://some.site/", headers={}, settings=settings)
        uri, options = api.prepare_request("/test", {"data": 123})
        assert uri == "http://some.site/test"
        assert options == {
            "headers": {
                "x-api-key": MOCK_API_KEY,
            },
            "data": 123,
        }

    def test_get_request_options(self, mocker):
        # substituting token avoids fetching/creating from config file
        MOCK_API_KEY = "mock_api_key"  # pragma: allowlist secret
        settings = Settings(auth=AuthSettings(api_key=MOCK_API_KEY))
        api = APIClient(host_uri=self.MOCK_URI, headers={}, settings=settings)
        options = api.get_request_options({"body": {"a": 1}})
        assert options["body"] == {"a": 1}

    def test_post(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch("covalent_cloud.shared.classes.api.requests.Session.__enter__", return_value=requests_mock)
    

        MOCK_REQUEST_OPTIONS = {"json": {"a": 1}, "data": b"ef8"}

        mocker.patch.object(api_mock, "get_request_options", return_value=MOCK_REQUEST_OPTIONS)
        api_mock.post("/api/v0/resource", MOCK_REQUEST_OPTIONS)
        requests_mock.post.assert_called_with(
            f"{self.MOCK_URI}/api/v0/resource", **MOCK_REQUEST_OPTIONS
        )

    def test_get(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch("covalent_cloud.shared.classes.api.requests.Session.__enter__", return_value=requests_mock)

        MOCK_REQUEST_OPTIONS = {
            "params": {"a": 1},
        }

        mocker.patch.object(api_mock, "get_request_options", return_value=MOCK_REQUEST_OPTIONS)
        api_mock.get("/api/v0/resource", MOCK_REQUEST_OPTIONS)
        requests_mock.get.assert_called_with(
            f"{self.MOCK_URI}/api/v0/resource", **MOCK_REQUEST_OPTIONS
        )

    def test_delete(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch("covalent_cloud.shared.classes.api.requests.Session.__enter__", return_value=requests_mock)

        mocker.patch.object(api_mock, "get_request_options")
        api_mock.delete("/api/v0/resource")
        requests_mock.delete.assert_called_with(f"{self.MOCK_URI}/api/v0/resource")

    def test_put(self, mocker):

        api_mock = APIClient(host_uri=self.MOCK_URI)
        requests_mock = MagicMock()
        mocker.patch("covalent_cloud.shared.classes.api.requests.Session.__enter__", return_value=requests_mock)

        MOCK_REQUEST_OPTIONS = {
            "json": {"a": 1},
        }

        mocker.patch.object(api_mock, "get_request_options", return_value=MOCK_REQUEST_OPTIONS)
        api_mock.put("/api/v0/resource", MOCK_REQUEST_OPTIONS)
        requests_mock.put.assert_called_with(
            f"{self.MOCK_URI}/api/v0/resource", **MOCK_REQUEST_OPTIONS
        )



# Copyright 2023 Agnostiq Inc.

from typing import Optional

from covalent_cloud.shared.classes.api import DeploymentAPI, DispatcherAPI
from covalent_cloud.shared.classes.settings import Settings, settings


def get_client(settings: Optional[Settings] = settings) -> DispatcherAPI:
    """
    A factory / getter method for the Dispatcher API client.

    Args:
        settings: An instance of `Settings` class. Default is `settings`.

    Returns:
        An instance of `DispatcherAPI` client.

    """

    return DispatcherAPI(settings=settings)


def get_deployment_client(settings: Optional[Settings] = settings) -> DeploymentAPI:
    """
    A factory / getter method for the Deployment API client.

    Args:
        settings: An instance of `Settings` class. Default is `settings`.

    Returns:
        An instance of `DeploymentAPI` client.

    """

    return DeploymentAPI(settings=settings)



# Copyright 2024 Agnostiq Inc.

import asyncio
import base64
import inspect
import json
from functools import partial
from typing import Any, List

import covalent as ct
from covalent._shared_files.context_managers import active_lattice_manager

from covalent_cloud.function_serve.common import (
    DEPLOY_ELECTRON_PREFIX,
    ServeAssetType,
    SupportedMethods,
    rename,
    wait_for_deployment_to_be_active,
)
from covalent_cloud.function_serve.models import Endpoint, FunctionServiceModel, ServeAsset
from covalent_cloud.shared.schemas.volume import Volume

__all__ = [
    "FunctionService",
]


TESTING_WARNING = "This is an unoptimized version of the app, useful only for local testing."
EXAMPLE_USAGE = "Sample usage to make a request to the endpoint: `requests.post('http://localhost:8000/your-route', params={'arg_name': 'arg_value'})`"
USE_JSON_WARNING = (
    "When using the deployed version, make sure to use `json=` instead of `params=` in the request"
)


class FunctionService:
    def __init__(
        self,
        func=None,
        executor=None,
        name=None,
        description=None,
        auth=None,
        tags=None,
        compute_share=None,
        volume=None,
        *,
        _main_func,
    ) -> None:
        self.init_func = func

        self.executor = executor

        self.func_name = None
        self.name = name
        self.func_source = None
        self.func_description = None

        if func is not None:
            self.func_name = func.__name__
            self.name = name or func.__name__
            self.func_source = inspect.getsource(func)
            self.func_description = (
                description
                or func.__doc__
                or "Add a docstring to your service function to populate this section."
            )

        self.auth = True if auth is None else auth
        self.tags = tags
        self.compute_share = compute_share

        self.all_endpoints: List[Endpoint] = []

        # This will be the main function that is used to start adding endpoints to
        # this will NEVER BE NONE
        self._main_func = _main_func

        # Set the volume to None by default, it will be set by the `cc.deploy` function
        self.volume: Volume = volume

        # Setting the root_dispatch_id to None by default - only valid for workflow-based deployments
        self.root_dispatch_id = None

    def _get_method_filtered_endpoints(self, method: str):
        return [ep for ep in self.all_endpoints if ep.method == method]

    def _is_testable_endpoint(self, func):
        sig = inspect.signature(func)

        # Return False if the function has any parameters without default values
        for param in sig.parameters.values():
            if param.default == inspect.Parameter.empty:
                return False

        return True

    def _method_decorator_generator(self, method, route, name, description, streaming, /, func):
        method_endpoints = self._get_method_filtered_endpoints(method)

        if route in [ep.route for ep in method_endpoints]:
            raise ValueError(f"A {method} method function already exists for this route")

        endpoint = Endpoint(
            route=route,
            name=name or func.__name__,
            endpoint_fn=ServeAsset(serialized_object=ct.TransportableObject(func).serialize()),
            endpoint_fn_source=inspect.getsource(func),
            streaming=streaming,
            description=description
            or func.__doc__
            or "Either add a docstring to your endpoint function or use the endpoint's 'description' parameter to populate this section.",
            test_endpoint_enabled=self._is_testable_endpoint(func),
            method=method,
        )
        # Register the route and the method function
        self.all_endpoints.append(endpoint)
        return endpoint

    # Assigning and returning the decorator like this as it is slightly easier to understand

    def _get(self, route: str, name: str = None, description: str = None, streaming: bool = False):
        get_decorator = partial(
            self._method_decorator_generator,
            SupportedMethods.GET,
            route,
            name,
            description,
            streaming,
        )
        return get_decorator

    def _post(
        self, route: str, name: str = None, description: str = None, streaming: bool = False
    ):
        post_decorator = partial(
            self._method_decorator_generator,
            SupportedMethods.POST,
            route,
            name,
            description,
            streaming,
        )
        return post_decorator

    def _put(self, route: str, name: str = None, description: str = None, streaming: bool = False):
        put_decorator = partial(
            self._method_decorator_generator,
            SupportedMethods.PUT,
            route,
            name,
            description,
            streaming,
        )
        return put_decorator

    def _delete(
        self, route: str, name: str = None, description: str = None, streaming: bool = False
    ):
        delete_decorator = partial(
            self._method_decorator_generator,
            SupportedMethods.DELETE,
            route,
            name,
            description,
            streaming,
        )
        return delete_decorator

    def _patch(
        self, route: str, name: str = None, description: str = None, streaming: bool = False
    ):
        patch_decorator = partial(
            self._method_decorator_generator,
            SupportedMethods.PATCH,
            route,
            name,
            description,
            streaming,
        )
        return patch_decorator

    def endpoint(
        self, route: str, name: str = None, description: str = None, streaming: bool = False
    ):
        return self._post(route, name, description, streaming)

    def _create_app(self, args, kwargs):
        """
        Create a FastAPI app from the instance of this class.
        FOR TESTING PURPOSES ONLY.
        """
        from fastapi import (
            FastAPI,  # NOTE: Importing here since Covalent Cloud doesn't have FastAPI as a dependency
        )

        # User warning to let them know that this is for testing purposes only
        print(TESTING_WARNING)
        print(EXAMPLE_USAGE)
        print(USE_JSON_WARNING)

        # Running the init function once before starting the app
        init_result = {}
        if self.init_func is not None:
            init_result = self.init_func(*args, **kwargs)

        print(f"Initial init result is: {init_result.keys()}")

        # Create a new FastAPI app from this instance
        app = FastAPI()

        # Add all the endpoints to the app
        for ep in self.all_endpoints:
            function_to = ct.TransportableObject.deserialize(ep.endpoint_fn.serialized_object)

            # Only attach those arguments which are also in the function signature of endpoint function
            func = function_to.get_deserialized()
            sig = inspect.signature(func)
            final_kwargs = {k: v for k, v in init_result.items() if k in sig.parameters}

            print(f"Final init result for endpoint {func.__name__} is: {final_kwargs.keys()}")

            attachable_fn = partial(function_to.get_deserialized(), **final_kwargs)

            if ep.method == SupportedMethods.GET:
                app.get(ep.route)(attachable_fn)
            elif ep.method == SupportedMethods.POST:
                app.post(ep.route)(attachable_fn)
            elif ep.method == SupportedMethods.PUT:
                app.put(ep.route)(attachable_fn)
            elif ep.method == SupportedMethods.DELETE:
                app.delete(ep.route)(attachable_fn)
            elif ep.method == SupportedMethods.PATCH:
                app.patch(ep.route)(attachable_fn)

        return app

    def test_run(self, port: int = 8000, *args, **kwargs):
        """
        Run the FastAPI app locally using uvicorn.
        This is useful for testing the FastAPI app before deploying it to a cloud service.
        """
        import uvicorn  # NOTE: Importing here since Covalent Cloud doesn't have uvicorn as a dependency

        # Run the app using uvicorn
        app = self._create_app(args, kwargs)

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop is not None:
            raise RuntimeError(
                f"Use `await {self.func_name}.test_run_async()` instead if there's a running event loop such as in Jupyter Notebook."
            )

        uvicorn.run(app, port=port)

    async def test_run_async(self, port: int = 8000, *args, **kwargs):
        """
        In case running inside a Jupyter notebook, await this method.

        Run the FastAPI app using uvicorn.
        This is useful for deploying the FastAPI app to a cloud service.
        """
        import uvicorn  # NOTE: Importing here since Covalent Cloud doesn't have uvicorn as a dependency

        app = self._create_app(args, kwargs)
        config = uvicorn.Config(app, port=port)
        server = uvicorn.Server(config)

        await server.serve()

    def get_model(self, *init_args, **init_kwargs):
        return FunctionServiceModel(
            title=self.name,
            description=self.func_description,
            executor=self.executor,
            compute_share=self.compute_share,
            tags=self.tags,
            auth=self.auth,
            init_fn=ServeAsset(
                serialized_object=ct.TransportableObject(self.init_func).serialize()
            ),
            init_fn_args=ServeAsset(
                serialized_object=ct.TransportableObject(list(init_args)).serialize()
            ),
            init_fn_args_json=ServeAsset(
                serialized_object=InitializerArgsEncoder.serialize(init_args),
                type=ServeAssetType.JSON,
            ),
            init_fn_kwargs=ServeAsset(
                serialized_object=ct.TransportableObject(init_kwargs).serialize()
            ),
            init_fn_kwargs_json=ServeAsset(
                serialized_object=InitializerArgsEncoder.serialize(init_kwargs),
                type=ServeAssetType.JSON,
            ),
            init_fn_source=self.func_source,
            endpoints=self.all_endpoints,
            volume=self.volume,
            root_dispatch_id=self.root_dispatch_id,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:

        import covalent_cloud as cc

        active_lattice = active_lattice_manager.get_active_lattice()
        if active_lattice is not None:

            dispatcher_uri = cc.settings.dispatcher_uri
            api_key = cc.get_api_key()

            @rename(DEPLOY_ELECTRON_PREFIX + self.func_name)
            def _electronic_deploy(
                function_service, dispatcher_uri, api_key, volume, *args, **kwargs
            ):

                import os

                import covalent_cloud as cc

                cc.settings.dispatcher_uri = dispatcher_uri
                cc.save_api_key(api_key)

                print("Current settings are: ")
                print(cc.settings.model_dump())

                root_dispatch_id = os.getenv("COVALENT_DISPATCH_ID")
                print("Root dispatch id is:", root_dispatch_id)
                function_service.root_dispatch_id = root_dispatch_id

                new_executor: cc.CloudExecutor = function_service.executor
                new_executor.settings = cc.settings.model_dump()
                function_service.executor = new_executor

                deployment = cc.deploy(function_service, volume)(*args, **kwargs)

                return wait_for_deployment_to_be_active(deployment, verbose=True)

            # Rename the deploy function to the name of the calling function
            deploy_func = partial(_electronic_deploy, self, dispatcher_uri, api_key, self.volume)
            deploy_func.__name__ = _electronic_deploy.__name__
            deploy_func.__doc__ = _electronic_deploy.__doc__
            deploy_func.__qualname__ = _electronic_deploy.__qualname__

            deploy_electron_executor = cc.CloudExecutor(
                env=self.executor.env, time_limit=60 * 60, num_cpus=4, memory=12 * 1024
            )
            # Create, run, and return the electron -> leveraging
            # the already built mechanisms in electrons
            return ct.electron(
                deploy_func,
                executor=deploy_electron_executor,
            )(*args, **kwargs)


class InitializerArgsEncoder(json.JSONEncoder):
    # A custom JSON encoder for converting the initializer function's
    # args and kwargs to JSON.

    # This encoder includes a fallback that defaults to encoding a string
    # representation of the object, meaning that recovering Python object
    # will not always be possible. This is fine because the matching pickled
    # objects are also available.

    # Nonetheless, typical initializer args and kwargs are expected to be JSON-serializable.
    def default(self, o):
        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            # fallback to string representation
            return json.dumps(str(o))

    @classmethod
    def serialize(cls, obj: Any) -> bytes:
        return base64.b64encode(json.dumps(obj, cls=cls).encode("utf-8"))

    @classmethod
    def deserialize(cls, ser_obj: bytes) -> Any:
        return json.loads(base64.b64decode(ser_obj).decode("utf-8"))



# Copyright 2023 Agnostiq Inc.

import uuid
from typing import Dict, List, Optional

from covalent._shared_files.qinfo import QNodeSpecs
from pydantic import BaseModel


class AssetMetadata(BaseModel):
    uri: str
    size: int


class CircuitAssets(BaseModel):
    circuit: Optional[AssetMetadata] = None
    circuit_string: Optional[AssetMetadata] = None
    result: Optional[AssetMetadata] = None
    result_string: Optional[AssetMetadata] = None


class QExecutorSpecs(BaseModel):
    name: str
    attributes: Dict


# Request body for `/register`
class CircuitCreateSchema(BaseModel):
    """Request body for the /register endpoint"""

    python_version: str
    dispatch_id: uuid.UUID
    node_id: int
    circuit_name: str
    circuit_description: str
    qnode_specs: QNodeSpecs

    # Clarify
    allowed_qexecutors: List[QExecutorSpecs]

    # Number of circuits in the batch
    num_circuits: int

    # Object store URIs -- assigned by the server
    assets: Optional[CircuitAssets] = None

    # Unique ID of a circuit -- assigned by the server
    circuit_id: Optional[str] = None

    selector: Optional[str] = None



# Copyright 2024 Agnostiq Inc.

import unittest
from contextlib import asynccontextmanager
from unittest.mock import patch

from covalent_cloud.function_serve.client import FunctionServeClient
from covalent_cloud.function_serve.common import ServiceStatus, wait_for_deployment_to_be_active
from covalent_cloud.function_serve.models import Deployment, RouteInfo

# NOTE: Abridged record. Needs to be kept up to date.
_DUMMY_FUNCTION_RECORD = {
    "id": "999999999999999",
    "title": "My LLM Service v1",
    "tags": ["llm", "llama2"],
    "invoke_url": "http://fake.covalent.xyz/fn/999999999999999",
    "auth": False,
    "description": "A fake LLM service for testing purposes.",
    "endpoints": [
        {
            "method": "POST",
            "route": "/infer",
            "name": "Non-streaming LLM Chatbot",
            "endpoint_fn": "<function_object_string>",
            "endpoint_fn_source": "def non_streaming(prompt, model):\n...",
            "streaming": False,
            "description": "<doc_string or user's description>",
            "test_endpoint_enabled": True,
        },
        {
            "method": "POST",
            "route": "/infer_streaming",
            "name": "Streaming LLM Chatbot",
            "endpoint_fn": "<function_object_string>",
            "endpoint_fn_source": "def streaming(prompt, model):\n...",
            "streaming": True,
            "description": "<doc_string or user's description>",
            "test_endpoint_enabled": True,
        },
    ],
    "status": "NEW_OBJECT",
}


class TestDeployment(unittest.IsolatedAsyncioTestCase):
    """Test the Deployment class."""

    def setUp(self) -> None:
        self.normal_route = RouteInfo(route="/custom123", method="POST", streaming=False)
        self.streaming_route = RouteInfo(
            route="/custom123_streaming", method="POST", streaming=True
        )
        self.deployment_id = "1234567890"
        self.address = "http://fake-address.com"
        self.name = "Fake Service for Testing"
        self.description = "A fake service for testing purposes."
        self.default_routes = [self.normal_route, self.streaming_route]
        self.status = ServiceStatus.ACTIVE
        self.tags = ["fake-tag", "fake-tag-2"]
        self.default_fsc = FunctionServeClient(
            function_id=self.deployment_id,
            host=self.address,
            token_getter=lambda: "fake-default-token",
        )

    def default_info(self):
        """Return a default Deployment object."""
        info = Deployment(
            function_id=self.deployment_id,
            address=self.address,
            name=self.name,
            description=self.description,
            routes=self.default_routes,
            status=self.status,
            tags=self.tags,
        )
        info.attach_route_methods(overwrite=True)
        return info

    def test_init_from_function_record(self):
        """Initialize a Deployment object with valid data."""

        fake_function_record = _DUMMY_FUNCTION_RECORD.copy()

        info = Deployment.from_function_record(fake_function_record)
        assert info.function_id == fake_function_record["id"]
        assert info.address == fake_function_record["invoke_url"]
        assert info.name == fake_function_record["title"]
        assert len(info.routes) == len(fake_function_record["endpoints"])
        assert info.status == fake_function_record["status"]
        assert info.tags == fake_function_record["tags"]

    def test_error_attribute(self):
        """Test that a deployment's `error` attribute is set correctly."""
        fake_fn_record_with_error = _DUMMY_FUNCTION_RECORD.copy()
        fake_fn_record_with_error["status"] = "ERROR"
        fake_fn_record_with_error[
            "error"
        ] = "Traceback (most recent call last):\nFake error message."

        info = Deployment.from_function_record(fake_fn_record_with_error)

        # Confirm error is assigned to `.error` attribute.
        assert info.error == fake_fn_record_with_error["error"]

        # Confirm error is rendered in the string representation.
        assert "───────── Exception ────────" in str(info)

        # Disable reload and check `wait_for_deployment_to_be_active` logic.
        with patch.object(Deployment, "reload", new=lambda self: None):
            for status in [ServiceStatus.ERROR, ServiceStatus.INACTIVE]:
                info.status = status
                info = wait_for_deployment_to_be_active(info)
                assert info.error == fake_fn_record_with_error["error"]

    def test_token_update(self):
        """Test that setting `info.token` uses the token in requests."""

        def fake_requests_request(*args, **kwargs):
            return FakeResponse(args, kwargs, self.default_info())

        # Patch to simply return call args, kwargs.
        with patch(
            "covalent_cloud.function_serve.client.requests.request",
            fake_requests_request,
        ):
            # Token is not set in `default_info`.
            info = self.default_info()
            response = info.custom123()
            assert "x-api-key" not in response["request_kwargs"]["headers"]

        with patch(
            "covalent_cloud.function_serve.client.requests.request",
            fake_requests_request,
        ):
            # Set token and retry.
            info.token = "fake-token"
            response = info.custom123()
            assert response["request_kwargs"]["headers"]["x-api-key"] == info.token

    def test_builtin_teardown_endpoint(self):
        """Test the builtin `teardown` method."""
        # Patch to do nothing instead of posting.
        with patch("requests.Session.put") as mock_put:
            with patch.object(
                FunctionServeClient, "teardown", wraps=self.default_fsc.teardown
            ) as mock_fsc_teardown:

                # Succeeds with a 'valid' API key.
                info = self.default_info()
                info.teardown()
                mock_fsc_teardown.assert_called_once()
                mock_put.assert_called_once()

    def test_builtin_info_endpoint(self):
        """Test the builtin `info` method."""

        def fake_requests_post(*args, **kwargs):
            return FakeResponse(args, kwargs, self.default_info())

        # Patch to simply return call args, kwargs.
        with patch(
            "covalent_cloud.function_serve.client.requests.post",
            fake_requests_post,
        ):
            with patch.object(
                FunctionServeClient, "info", wraps=self.default_fsc.info
            ) as mock_fsc_info:

                info = self.default_info()
                response = info.info()
                mock_fsc_info.assert_called_once()
                assert response["request_args"][0] == f"{info.address}/info"

    def test_user_endpoint_sync(self):
        """
        Test requests posted to user-defined service endpoints.
        Synchronous, non-streaming requests.
        """

        def fake_requests_post(*args, **kwargs):
            return FakeResponse(args, kwargs, self.default_info())

        # Patch to simply return call args, kwargs.
        with patch(
            "covalent_cloud.function_serve.client.requests.request",
            fake_requests_post,
        ):
            with patch.object(
                FunctionServeClient,
                "_request",
                wraps=self.default_fsc._request,  # pylint: disable=protected-access
            ) as mock_fsc_request:

                info = self.default_info()
                response = info.custom123(some_param="some_value", another_param="another_value")
                mock_fsc_request.assert_called_once()
                assert response["request_kwargs"]["method"] == self.normal_route.method
                assert (
                    response["request_kwargs"]["url"] == f"{info.address}{self.normal_route.route}"
                )
                assert response["request_kwargs"]["json"] == {
                    "some_param": "some_value",
                    "another_param": "another_value",
                }
                assert response["request_kwargs"]["timeout"] is None

    async def test_user_endpoint_async(self):
        """
        Test requests posted to user-defined service endpoints.
        Asynchronous, streaming requests.
        """
        # Patch to simply return call args, kwargs.
        with patch(
            "covalent_cloud.function_serve.client.aiohttp.ClientSession",
            AsyncFakeClientSession,
        ):
            with patch.object(
                FunctionServeClient,
                "_async_request",
                wraps=self.default_fsc._async_request,  # pylint: disable=protected-access
            ) as mock_fsc_async_request:

                info = self.default_info()
                response = await info.async_custom123(
                    some_param="some_value", another_param="another_value"
                )
                mock_fsc_async_request.assert_called_once()
                assert response["request_kwargs"]["method"] == self.normal_route.method
                assert (
                    response["request_kwargs"]["url"] == f"{info.address}{self.normal_route.route}"
                )
                assert response["request_kwargs"]["json"] == {
                    "some_param": "some_value",
                    "another_param": "another_value",
                }

    def test_user_endpoint_streaming_sync(self):
        """
        Test requests posted to user-defined service endpoints.
        Synchronous, streaming requests.
        """

        def fake_requests_post(*args, **kwargs):
            return FakeResponse(args, kwargs, self.default_info())

        # Patch to simply return call args, kwargs.
        with patch(
            "covalent_cloud.function_serve.client.requests.request",
            fake_requests_post,
        ):
            with patch.object(
                FunctionServeClient,
                "_request_streaming",
                wraps=self.default_fsc._request_streaming,  # pylint: disable=protected-access
            ) as mock_fsc_request_streaming:

                info = self.default_info()
                info.token = "fake-token"
                words = []
                for word in info.custom123_streaming(
                    some_param="some_value", another_param="another_value"
                ):
                    words.append(word)
                mock_fsc_request_streaming.assert_called_once()

            assert " ".join(words) == FakeResponse.fake_content

    async def test_user_endpoint_streaming_async(self):
        """
        Test requests posted to user-defined service endpoints.
        Asynchronous, streaming requests.
        """
        # Patch to simply return call args, kwargs.
        with patch(
            "covalent_cloud.function_serve.client.aiohttp.ClientSession",
            FakeClientSession,
        ):
            with patch.object(
                FunctionServeClient,
                "_async_request_streaming",
                wraps=self.default_fsc._async_request_streaming,  # pylint: disable=protected-access
            ) as mock_fsc_async_request_streaming:

                info = self.default_info()
                info.token = "fake-token"
                words = []
                async for word in info.async_custom123_streaming(
                    some_param="some_value", another_param="another_value"
                ):
                    words.append(word)
                mock_fsc_async_request_streaming.assert_called_once()

            assert " ".join(words) == FakeResponse.fake_content


class FakeClientSession:
    """Stand-in for an `aiohttp.ClientSession` object. Avoid making real requests."""

    @asynccontextmanager
    async def request(self, *args, **kwargs):
        """Return a fake response object."""
        try:
            t_obj = TestDeployment()
            t_obj.setUp()
            yield FakeResponse(args, kwargs, info=t_obj.default_info())
        finally:
            pass

    async def __aenter__(self):
        """Return self for use in async context managers."""
        return self

    async def __aexit__(self, *args):
        """Do nothing when exiting async context managers."""
        pass


class AsyncFakeClientSession(FakeClientSession):

    @asynccontextmanager
    async def request(self, *args, **kwargs):
        try:
            t_obj = TestDeployment()
            t_obj.setUp()
            yield AsyncFakeResponse(args, kwargs, info=t_obj.default_info())
        finally:
            pass


class FakeResponse:
    """Stand-in for a `requests.Response` object. Avoid makings real requests."""

    fake_content = "This is a fake generated response from a fake LLM."

    @property
    def content(self):
        """Return a fake response body."""
        return FakeContent(self.fake_content)

    def __init__(self, args, kwargs, info: Deployment, error=False):
        self.request_args = args
        self.request_kwargs = kwargs
        self.info = info
        self.error = error

    @property
    def text(self):
        return self.fake_content

    def raise_for_status(self):
        """Raise a fake exception if necessary."""
        if self.error:
            raise Exception("Fake error")

    def iter_content(self):
        """Iterate over a fake response body."""
        for word in self.fake_content.split():
            yield word

    def json(self):
        """Return a fake response body."""
        d = self.info.model_dump()
        name_map = {
            "id": "function_id",
            "title": "name",
            "invoke_url": "address",
            "endpoints": "routes",
        }
        for m, n in name_map.items():
            d[m] = d.pop(n)
        d["request_args"] = self.request_args
        d["request_kwargs"] = self.request_kwargs
        return d


class AsyncFakeResponse(FakeResponse):

    async def text(self):
        return self.fake_content

    async def json(self):
        return super().json()


class FakeContent:
    """Stand-in for an `aiohttp` response content object."""

    def __init__(self, content):
        self.content = content

    async def iter_any(self):
        """Iterate over a fake response body."""
        for word in self.content.split():
            yield word



# Copyright 2024 Agnostiq Inc.

from functools import partial
from typing import Any, AsyncGenerator, Callable, Dict, Generator

import aiohttp
import requests
from aiohttp import ClientResponse
from requests import Response

from covalent_cloud.service_account_interface.auth_config_manager import get_api_key
from covalent_cloud.service_account_interface.client import get_deployment_client

__all__ = ["FunctionServeClient"]


MSG_HOST_NONE = "Address is None. If the deployment status is not ERROR, try `deployment.reload()` or wait for the deployment to be active."


class FunctionServeClient:
    """
    Client for users to interact with a deployed Function Serve API.
    """

    def __init__(self, function_id: str, host: str, token_getter: Callable) -> None:
        """
        Initialize the client.

        Args:
            host: The host of the deployed service.
            token_getter: A function that returns the service auth token.
        """
        self.function_id = function_id
        self.host = host
        self.token_getter = token_getter

    def get_headers(self, **kwargs) -> Dict[str, str]:
        """
        Get the headers for the request.
        """
        headers = {}

        if token := kwargs.pop("token", None):
            headers["x-api-key"] = token
        if api_key := kwargs.pop("api_key", None):
            headers["x-api-key"] = api_key

        return headers

    def teardown(self) -> Response:
        """
        Permanently tear down the deployment.
        Special case - posts to Covalent Function Service (CFS) API.
        Requires Covalent Cloud API key, if key is not saved in local config.
        """
        cfs_api_client = get_deployment_client()
        response = cfs_api_client.put(f"/functions/{self.function_id}/teardown")
        try:
            res = response.json()
            response.raise_for_status()
        except Exception as e:
            print("Response:", response.text)
            raise e
        return res

    def info(self) -> Response:
        """
        Get the deployment info.
        """
        url = f"{self.host}/info"
        headers = self.get_headers(api_key=get_api_key())
        response = requests.post(url, headers=headers, timeout=10)
        try:
            res = response.json()
            response.raise_for_status()
        except Exception as e:
            print("Response:", response.text)
            raise e
        return res

    def make_request_method(
        self,
        route: str,
        method: str,
        streaming: bool,
        is_async: bool,
    ) -> Callable:
        """
        Get a partial function for making requests to a specific route.
        """
        route = route if route.startswith("/") else f"/{route}"

        if streaming and is_async:
            func = partial(self._async_request_streaming, route, method)
        elif streaming:
            func = partial(self._request_streaming, route, method)
        elif is_async:
            func = partial(self._async_request, route, method)
        else:
            func = partial(self._request, route, method)
        return func

    # Generic HTTP request methods for above partials.

    def _request(
        self,
        route: str,
        method: str,
        **data: Any,
    ) -> Response:
        """
        Make synchronous requests to a url.
        """

        if self.host is None or self.host == "None":
            raise ValueError(MSG_HOST_NONE)

        url = f"{self.host}{route}"
        headers = self.get_headers(token=self.token_getter())
        request_kwargs = {
            "method": method,
            "url": url,
            "headers": headers,
            "json": data or {},
            "timeout": None,
        }
        response = requests.request(**request_kwargs)

        try:
            res = response.json()
            response.raise_for_status()
        except Exception as e:
            print("Response:", response.text)
            raise e

        return res

    async def _async_request(
        self,
        route: str,
        method: str,
        **data: Any,
    ) -> ClientResponse:
        """
        Make asynchronous requests to a url.
        """

        if self.host is None or self.host == "None":
            raise ValueError(MSG_HOST_NONE)

        url = f"{self.host}{route}"
        headers = self.get_headers(token=self.token_getter())
        request_kwargs = {
            "method": method,
            "url": url,
            "headers": headers,
            "json": data or {},
        }
        async with aiohttp.ClientSession() as session:
            async with session.request(**request_kwargs) as response:
                try:
                    res = await response.json()
                    response.raise_for_status()
                except Exception as e:
                    print("Response:", await response.text())
                    raise e

                return res

    def _request_streaming(
        self,
        route: str,
        method: str,
        **data: Any,
    ) -> Generator:
        """
        Make synchronous streaming requests to a url.
        """

        if self.host is None or self.host == "None":
            raise ValueError(MSG_HOST_NONE)

        url = f"{self.host}{route}"
        headers = self.get_headers(token=self.token_getter())
        request_kwargs = {
            "method": method,
            "url": url,
            "headers": headers,
            "json": data or {},
            "stream": True,
            "timeout": None,
        }
        response = requests.request(**request_kwargs)

        try:
            response.raise_for_status()
        except Exception as e:
            print("Response:", response.text)
            raise e

        for chunk in response.iter_content():
            yield chunk

    async def _async_request_streaming(
        self,
        route: str,
        method: str,
        **data: Any,
    ) -> AsyncGenerator:
        """
        Make asynchronous streaming requests to a url.
        """

        if self.host is None:
            raise ValueError()

        url = f"{self.host}{route}"
        headers = self.get_headers(token=self.token_getter())
        request_kwargs = {
            "method": method,
            "url": url,
            "headers": headers,
            "json": data or {},
        }
        async with aiohttp.ClientSession() as session:
            async with session.request(**request_kwargs) as response:

                try:
                    response.raise_for_status()
                except Exception as e:
                    print("Response:", await response.text())
                    raise e

                async for chunk in response.content.iter_any():
                    yield chunk



# Copyright 2023 Agnostiq Inc.


from enum import Enum


class GPU_TYPE(Enum):
    """
    Enum for GPU types.
    """

    def __str__(self):
        return self.value

    A100 = "a100-80g"
    H100 = "h100"
    A4000 = "a4000"
    A5000 = "a5000"
    A6000 = "a6000"
    A10 = "a10"
    T4 = "t4"
    L40 = "l40"
    V100 = "v100"
    REMOTELY_SET = "remotely_set"


DEPRECATED_GPUS = [GPU_TYPE.V100, GPU_TYPE.A4000, GPU_TYPE.A5000]
DEPRECATED_GPUS.extend([str(gpu) for gpu in DEPRECATED_GPUS])



# Copyright 2025 Agnostiq Inc.

import sys
from unittest.mock import patch

import pytest

import covalent_cloud as cc
from covalent_cloud.shared.classes.api import DispatcherAPI
from covalent_cloud.swe_management.swe_manager import PY_VERSION_WARNING_TEMPLATE

_MOCK_LOCAL_VERSION_INFO = (3, 10, 16)
_MOCK_LOCAL_VERSION = ".".join(map(str, _MOCK_LOCAL_VERSION_INFO))


def test_python_version_unspecified():
    """Test that no warning is raised when Python version is not specified."""
    with patch.object(DispatcherAPI, "post"):
        with pytest.warns(None):
            cc.create_env("testing_env", pip=[], conda=[], wait=False)


def test_python_version_exact():
    """Test that the exact local Python version is not warned against."""
    with patch.object(DispatcherAPI, "post"), patch.object(
        sys, "version_info", _MOCK_LOCAL_VERSION_INFO
    ):
        with pytest.warns(None):
            cc.create_env(
                "testing_env", pip=[], conda=[f"python={_MOCK_LOCAL_VERSION}"], wait=False
            )


@pytest.mark.parametrize(
    "req_version",
    [
        # Strict requirement, different from local version.
        ("python=3.1.2"),
        ("python=3.9.1"),
        ("python=2.0.8"),
        ("python=3.8.4"),
        ("python=3.9"),
        ("python=3.7"),
        ("python=3"),
        # Flexible requirement with bound at same as local version.
        (f"python<{_MOCK_LOCAL_VERSION}"),
        (f"python>{_MOCK_LOCAL_VERSION}"),
        (f"python<={_MOCK_LOCAL_VERSION}"),
        (f"python>={_MOCK_LOCAL_VERSION}"),
        (f"python!={_MOCK_LOCAL_VERSION}"),
        # Multi-part requirement, necessarily flexible.
        ("python>=3.9,<3.12"),
        ("python>=3.9,<3.12,!=3.9.2"),
    ],
)
def test_python_version_warnings(req_version):
    """Test valid Python versions that cause a warning due to local environment."""
    with patch.object(DispatcherAPI, "post"):
        with pytest.warns(UserWarning) as record:
            cc.create_env("testing_env", pip=[], conda=[req_version], wait=False)

    assert len(record) == 1
    assert record[0].message.args[0] == PY_VERSION_WARNING_TEMPLATE.format(
        local_version=_MOCK_LOCAL_VERSION,
        dependency=req_version,
    )


@pytest.mark.parametrize(
    "req_version",
    [
        ("python 3.10"),
        ("python"),
        ("python::3.10"),
    ],
)
def test_python_dependency_invalid(req_version):
    """
    Test failure to match the Python version regex.
    Cases must be recognized as Python versions to be considered.
    E.g. 'pythoon=3.10' will not raise an error here.
    """
    with patch.object(DispatcherAPI, "post"):
        with pytest.raises(ValueError, match="Invalid python dependency: "):
            cc.create_env("testing_env", pip=[], conda=[req_version], wait=False)


@pytest.mark.parametrize(
    "req_version",
    [
        ("python>=3,<-312"),
        ("python>3,abcde"),
        ("python="),
    ],
)
def test_python_version_invalid(req_version):
    """
    Test failure due to `packaging.version.parse` errors.
    These errors happen when the pattern is matched but the
    version code is not parsable.
    """
    with patch.object(DispatcherAPI, "post"):
        with pytest.raises(ValueError, match="Invalid python version: "):
            cc.create_env("testing_env", pip=[], conda=[req_version], wait=False)



# Copyright 2023 Agnostiq Inc.


import contextlib

import requests
from rich.console import Console

from covalent_cloud.shared.classes.settings import settings

"""Covalent Cloud SDK Exception module."""


class CovalentSDKError(Exception):
    """Covalent Cloud SDK Base Exception class.

    Attributes:
        message (str): Explanation of the error.
        code (str): String enum representing error analogous to error code.
    """

    def __init__(self, message: str = "Generic Error", code: str = "error/generic") -> None:
        """
        Initializes a new instance of the CovalentSDKError class.

        Args:
            message (str): Explanation of the error.
            code (str): String enum representing error analogous to error code.

        """
        self.message = message
        self.code = code
        super().__init__(f"[{code}] {message}")

    @staticmethod
    def print_error(e: Exception, level: str = "warning") -> None:
        """Print a CovalentSDKError.

        Args:
            e: The CovalentSDKError to print.
            level: The level of the message to print.  Defaults to "warning".

        """
        message = str(e)

        # if it is an error raised by APIClient (requests lib under the hood) parse out error message and display
        if isinstance(e, requests.HTTPError):
            error_response = e.response
            with contextlib.suppress(Exception):
                message = error_response.json().get("detail", message)

        console = Console()
        if level == "warning":
            console.print(f"[bold yellow1]WARNING: {message}[bold yellow1]")
        elif level == "error":
            console.print(f"[bold red1]ERROR: {message}[bold red1]")

    def rich_print(self, level: str = "warning") -> None:
        CovalentSDKError.print_error(self, level)


class CovalentAPIKeyError(CovalentSDKError):
    """Covalent Cloud SDK API Key Error class."""

    def __init__(self, message, code) -> None:
        super().__init__(message, code)


class CovalentGenericAPIError(CovalentSDKError):
    """Covalent Cloud Server Generic API Error class."""

    def __init__(self, error) -> None:
        try:
            error_message = error.response.json()["detail"]
            error_code = error.response.json()["code"]
        except:
            error_message = "Unknown Error"
            error_code = "error/unknown"

        super().__init__(error_message, error_code)


class InsufficientMemoryError(CovalentSDKError):
    """Error raised when the assigned executor's memory is insufficient for reliable service operation."""

    def __init__(self, memory) -> None:
        message = f"The assigned executor's memory ({memory} GB) is insufficient for reliable service operation. Assign an executor with more than {settings.function_serve.min_executor_memory_gb} GB of memory to avoid this error (a minimum of {settings.function_serve.recommended_executor_memory_gb} GB is recommended)."
        code = "error/insufficient_memory"
        super().__init__(message, code)


def handle_error(e):
    """Handle a Covalent Cloud SDK Error.

    Args:
        e: The Covalent Cloud SDK Error to handle.

    """
    if isinstance(e, CovalentSDKError):
        e.rich_print(level="error")
    else:
        CovalentSDKError.rich_print(e, level="error")



# Copyright 2024 Agnostiq Inc.

import inspect
import typing
from typing import Any, Callable, Union

from covalent_cloud.function_serve.assets import AssetsMediator
from covalent_cloud.function_serve.common import wait_for_deployment_to_be_active
from covalent_cloud.function_serve.models import Deployment
from covalent_cloud.service_account_interface.client import get_deployment_client
from covalent_cloud.shared.classes.settings import Settings, settings
from covalent_cloud.shared.schemas.volume import Volume

if typing.TYPE_CHECKING:
    from covalent_cloud.function_serve.service_class import FunctionService

__all__ = [
    "deploy",
    "get_deployment",
]


def deploy(
    function_service: "FunctionService", volume: Volume = None, settings: Settings = settings
) -> Callable[[Any], Deployment]:
    """Deploy a function service to Covalent Cloud.

    Args:
        function_service: A function decorated with `@cc.service`.
        volume: Grant access to a cloud storage volume in Covalent Cloud. Defaults to None.
        settings: User settings for Covalent Cloud. Defaults to settings on the client machine.

    Returns:
        A callable which launches the deployment and has the same signature as the initializer
        for `function_service`.
    """

    def deploy_wrapper(*args, **kwargs) -> Deployment:

        # Force a TypeError if the arguments are invalid.
        # If not done here, error will be raised in remote host.
        sig = inspect.signature(function_service.init_func)
        sig.bind(*args, **kwargs)

        if volume is not None:
            # Override the volume for the function service
            function_service.volume = volume

        fn_service_model = function_service.get_model(*args, **kwargs)

        assets_mediator = AssetsMediator()
        fn_service_model = assets_mediator.hydrate_assets_from_model(
            fn_service_model, settings=settings
        )

        assets_mediator.upload_all()

        dumped_model = fn_service_model.model_dump()

        deployment_client = get_deployment_client(settings)
        response = deployment_client.post(
            "/functions",
            request_options={
                "json": dumped_model,
            },
        )

        deployment = Deployment.from_function_record(response.json())

        # Attach route methods for ease of use
        deployment.attach_route_methods()

        return deployment

    return deploy_wrapper


def get_deployment(
    function_id: Union[str, Deployment],
    wait: Union[bool, int, float] = False,
    settings: Settings = settings
) -> Deployment:
    """Retrieve or refresh a client object for a deployed function service.

    Args:
        function_id: ID string or client object for the target deployment.
        wait: Option to wait for the deployment to be active. Defaults to False.
            Numerical values represent the approximate time to wait (in seconds)
            for the deployment to finish initializing, before raising a client-side `TimeoutError`.
            The boolean value True corresponds to 3600, i.e. 1 hour.
        settings: User settings for Covalent Cloud. Defaults to settings on the client machine.

    Returns:
        Deployment: Deployment object for the function service.
    """

    if isinstance(function_id, Deployment):
        function_id = function_id.function_id

    deployment_client = get_deployment_client(settings)
    response = deployment_client.get(f"/functions/{function_id}")
    deployment = Deployment.from_function_record(response.json())

    # Attach route methods for ease of use
    deployment.attach_route_methods()

    if not wait:
        return deployment

    if isinstance(wait, bool):
        wait_time_max = 3600
    elif isinstance(wait, (int, float)) and wait > 0:
        wait_time_max = wait
    else:
        raise ValueError("Invalid value for `wait`. Must be a boolean or a positive int or float.")

    return wait_for_deployment_to_be_active(deployment, wait_time_max=wait_time_max)
